import logging
from abc import ABC, abstractmethod
from subprocess import CalledProcessError
from typing import TypeVar, Generic

from ddb.action import Action, InitializableAction
from ddb.config import config
from ddb.context import context
from ddb.context.context import ContextStackItem

A = TypeVar('A', bound=Action)  # pylint:disable=invalid-name


class FailFastError(Exception):
    """
    A base exception that should always fail fast, even when flag is not enabled.
    """


class ExpectedError(Exception):
    """
    A base exception that can display it's own message in logs.
    """

    def log_error(self):
        """
        Log the error
        :return:
        """
        log_error = context.log.exception if \
            context.log.isEnabledFor(logging.DEBUG) or config.args.exceptions \
            else context.log.error
        log_error(str(self))


# pylint:disable=too-few-public-methods


class EventBindingRunner(Generic[A], ABC):
    """
    Runner for an event binding.
    """

    def __init__(self, action: A, event_name: str, to_call=None):
        self.action = action
        self.event_name = event_name
        self.to_call = to_call if to_call else action.execute

    @abstractmethod
    def run(self, *args, **kwargs):
        """
        Run the callable registered on this event binding.
        """


class ActionEventBindingRunner(Generic[A], EventBindingRunner[A]):
    """
    Runner for an action event binding.
    """

    def __init__(self, action: A, event_name: str, to_call=None, event_processor=None, fail_fast=False):
        super().__init__(action, event_name, to_call)
        self.event_processor = event_processor
        self.fail_fast = fail_fast

    def run(self, *args, **kwargs):  # pylint:disable=missing-function-docstring
        if self.event_processor:
            resp = self.event_processor(*args, **kwargs)
            if resp is False or resp is None:
                return None
            try:
                args, kwargs = resp
            except ValueError:
                pass

        context.stack.append(ContextStackItem(self.event_name, self.action, self.to_call, args, kwargs))
        try:
            if context.log.isEnabledFor(logging.DEBUG):
                context.log.debug("Execute action %s", context.stack)

            self._execute_action(*args, **kwargs)
            return True
        except Exception as exception:  # pylint:disable=broad-except
            self._handle_exception(exception)
            if self.fail_fast or isinstance(exception, FailFastError):
                raise
            return False
        finally:
            if context.stack:
                context.stack.pop()

    def _execute_action(self, *args, **kwargs):
        """
        Execute the action
        """
        self.to_call(*args, **kwargs)

    @staticmethod
    def _handle_exception(exception: Exception):
        """
        Handle an exception occured in an action.
        """
        context.exceptions.append(exception)
        log_error = context.log.exception if \
            context.log.isEnabledFor(logging.DEBUG) or config.args.exceptions \
            else context.log.error
        if isinstance(exception, ExpectedError):
            exception.log_error()
        else:
            log_error("An unexpected error has occured %s: %s",
                      context.stack,
                      str(exception).strip())
        if isinstance(exception, CalledProcessError):
            if exception.stderr:
                for err_line in exception.stderr.decode("utf-8").splitlines():
                    log_error("(stderr) %s", err_line)
            else:
                for out_line in exception.stderr.decode("utf-8").splitlines():
                    log_error("(output) %s", out_line)


class InitializableActionEventBindingRunner(ActionEventBindingRunner[InitializableAction]):
    """
    Runner for an initializable action event binding
    """

    def run(self, *args, **kwargs):
        """
        Initialize if required and run
        """
        if not self.action.initialized:
            self.action.initialize()
            self.action.initialized = True
        return super().run(*args, **kwargs)



# -*- coding: utf-8 -*-
import json
import os
import re
import shlex
from collections import namedtuple
from subprocess import CalledProcessError
from typing import Iterable

from chmod_monkey import tmp_chmod
from dockerfile_parse import DockerfileParser
from dotty_dict import Dotty

from ..copy.actions import copy_from_url
from ...action import Action
from ...action.action import EventBinding
from ...config import config
from ...context import context
from ...event import events
from ...utils.file import force_remove
from ...utils.process import run

BuildServiceDef = namedtuple("BuildServiceDef", "context dockerfile")


class CustomDockerfileParser(DockerfileParser):
    """
    Custom class to implement entrypoint property with the same behavior as cmd property.
    """

    def get_last_instruction(self, instruction_type, instruction_condition=lambda instruction: True):
        """
        Determine the final instruction_type instruction, if any, in the final build stage.
        instruction_types from earlier stages are ignored.
        :return: value of final stage instruction_type instruction
        """
        last_instruction = None
        for insndesc in self.structure:
            if insndesc['instruction'] == 'FROM':  # new stage, reset
                last_instruction = None
            elif insndesc['instruction'] == instruction_type and instruction_condition(insndesc):
                last_instruction = insndesc
        return last_instruction

    @property
    def entrypoint(self):
        """
        Determine the final ENTRYPOINT instruction, if any, in the final build stage.
        ENTRYPOINTs from earlier stages are ignored.
        :return: value of final stage ENTRYPOINT instruction
        """
        last_instruction = self.get_last_instruction("ENTRYPOINT")
        return last_instruction['value'] if last_instruction else None

    @entrypoint.setter
    def entrypoint(self, value):
        """
        setter for final 'ENTRYPOINT' instruction in final build stage
        """
        entrypoint = None
        for insndesc in self.structure:
            if insndesc['instruction'] == 'FROM':  # new stage, reset
                entrypoint = None
            elif insndesc['instruction'] == 'ENTRYPOINT':
                entrypoint = insndesc

        if value:
            new_entrypoint = 'ENTRYPOINT ' + value
        else:
            new_entrypoint = None

        if entrypoint:
            if new_entrypoint:
                self.add_lines_at(entrypoint, new_entrypoint, replace=True)
            else:
                new_lines = list(self.lines)
                new_lines.remove(entrypoint['content'])
                self.lines = new_lines
        elif new_entrypoint:
            self.add_lines(new_entrypoint)

    @property
    def cmd(self):
        """
        Determine the final CMD instruction, if any, in the final build stage.
        CMDs from earlier stages are ignored.
        :return: value of final stage CMD instruction
        """
        value = None
        for insndesc in self.structure:
            if insndesc['instruction'] == 'FROM':  # new stage, reset
                value = None
            elif insndesc['instruction'] == 'CMD':
                value = insndesc['value']
        return value

    @cmd.setter
    def cmd(self, value):
        """
        setter for final 'CMD' instruction in final build stage

        """
        cmd = None
        for insndesc in self.structure:
            if insndesc['instruction'] == 'FROM':  # new stage, reset
                cmd = None
            elif insndesc['instruction'] == 'CMD':
                cmd = insndesc

        if value:
            new_cmd = 'CMD ' + value
        else:
            new_cmd = None
        if cmd:
            if new_cmd:
                self.add_lines_at(cmd, new_cmd, replace=True)
            else:
                new_lines = list(self.lines)
                new_lines.remove(cmd['content'])
                self.lines = new_lines
        elif new_cmd:
            self.add_lines(new_cmd)


class FixuidDockerComposeAction(Action):
    """
    Automate fixuid configuration for docker compose services where fixuid.yml configuration file is available in
    build context
    """

    def __init__(self):
        self.docker_compose_config = {}
        self._dockerfile_lines = ("ADD fixuid.tar.gz /usr/local/bin",
                                  "RUN chown root:root /usr/local/bin/fixuid && "
                                  "chmod 4755 /usr/local/bin/fixuid && "
                                  "mkdir -p /etc/fixuid",
                                  "COPY fixuid.yml /etc/fixuid/config.yml")

    @property
    def event_bindings(self):
        def file_generated_processor(source: str, target: str):
            service = self.find_fixuid_service(target)
            if service:
                return (), {"service": service}
            return None

        def file_found_processor(file: str):
            if os.path.basename(file) == 'fixuid.yml':
                dockerfile_filepath = os.path.join(os.path.dirname(file), 'Dockerfile')
                service = self.find_fixuid_service(dockerfile_filepath)
                if service:
                    return (), {"service": service}
            return None

        def file_deleted_processor(file: str):
            if os.path.basename(file) == 'fixuid.yml':
                dockerfile_filepath = os.path.join(os.path.dirname(file), 'Dockerfile')
                service = self.find_fixuid_service(dockerfile_filepath, include_missing_fixuid=True)
                if service:
                    return (), {"service": service}
            return None

        return (
            events.docker.docker_compose_config,
            EventBinding(events.file.generated,
                         call=self.apply_fixuid,
                         processor=file_generated_processor),
            EventBinding(events.file.found,
                         call=self.apply_fixuid,
                         processor=file_found_processor),
            EventBinding(events.file.deleted,
                         call=self.remove_fixuid,
                         processor=file_deleted_processor)
        )

    @property
    def name(self) -> str:
        return "fixuid:docker"

    @staticmethod
    def _get_inspect_data(image):
        try:
            context.log.debug("Inspect base image %s", image)
            output = run('docker', 'inspect', '--format', '\'{{json .}}\'', image)
        except CalledProcessError:
            context.log.warning("Pull base image %s (downloading)", image)
            run('docker', 'pull', '-q', image)
            context.log.success("Pull base image %s (terminated)", image)
            context.log.debug("Inspect base image %s", image)
            output = run('docker', 'inspect', '--format', '\'{{json .}}\'', image)
        output = output.decode().strip('\\\'\n').rstrip('\\\'\n')
        inspect_data = json.loads(output)

        context.log.debug("%s (%s)", image, inspect_data['Id'])
        return inspect_data

    @staticmethod
    def _get_image_config(image):
        if image and image != 'scratch':
            attrs = FixuidDockerComposeAction._get_inspect_data(image)
            if attrs and 'Config' in attrs:
                return attrs['Config']
        return None

    def apply_fixuid(self, service: BuildServiceDef):
        """
        Apply fixuid to given service
        """
        dockerfile_path = os.path.join(service.context, service.dockerfile)
        if os.path.exists(dockerfile_path):
            with tmp_chmod(dockerfile_path, '+w'):
                with open(dockerfile_path, "ba+") as dockerfile_file:
                    parser = CustomDockerfileParser(fileobj=dockerfile_file)

                    if FixuidDockerComposeAction._apply_fixuid_from_parser(self, parser, service):
                        context.log.success("Fixuid applied to %s",
                                            os.path.relpath(dockerfile_path, config.paths.project_home))

    def remove_fixuid(self, service: BuildServiceDef):
        """
        Remove fixuid from given service
        """
        dockerfile_path = os.path.join(service.context, service.dockerfile)
        if os.path.exists(dockerfile_path):
            with tmp_chmod(dockerfile_path, '+w'):
                with open(dockerfile_path, "ba+") as dockerfile_file:
                    parser = CustomDockerfileParser(fileobj=dockerfile_file)

                    if FixuidDockerComposeAction._remove_fixuid_from_parser(self, parser, service):
                        context.log.success("Fixuid removed from %s",
                                            os.path.relpath(dockerfile_path, config.paths.project_home))

    @staticmethod
    def _get_cmd_and_entrypoint(parser):
        entrypoint = parser.entrypoint
        cmd = parser.cmd
        # if entrypoint is defined in Dockerfile, we should not grab cmd from base image
        # and reset CMD to empty value to stay consistent with docker behavior.
        # see https://github.com/docker/docker.github.io/issues/6142
        reset_cmd = False
        if entrypoint and not cmd:
            reset_cmd = True
        if not entrypoint:
            baseimage_config = FixuidDockerComposeAction._get_image_config(parser.baseimage)
            if baseimage_config and 'Entrypoint' in baseimage_config:
                entrypoint = json.dumps(baseimage_config['Entrypoint'])
        if not cmd and not reset_cmd:
            baseimage_config = FixuidDockerComposeAction._get_image_config(parser.baseimage)
            if baseimage_config and 'Cmd' in baseimage_config:
                cmd = json.dumps(baseimage_config['Cmd'])
        if not cmd:
            cmd = None
        return cmd, entrypoint

    @staticmethod
    def _has_fixuid_disabled_comment(lines: Iterable[str]):
        return FixuidDockerComposeAction._has_comment(
            ['no-fixuid', 'fixuid-no', 'fixuid-disabled?', 'disabled?-fixuid'], lines)

    @staticmethod
    def _has_fixuid_manual_comment(lines: Iterable[str]):
        return FixuidDockerComposeAction._has_comment(
            ['manual-fixuid', 'fixuid-manual', 'custom-fixuid', 'fixuid-custom'], lines)

    @staticmethod
    def _has_fixuid_manual_entrypoint_comment(lines: Iterable[str]):
        return FixuidDockerComposeAction._has_comment(
            ['manual-fixuid-entrypoint',
             'fixuid-manual-entrypoint',
             'custom-fixuid-entrypoint',
             'fixuid-custom-entrypoint'], lines)

    @staticmethod
    def _has_fixuid_manual_install_comment(lines: Iterable[str]):
        return FixuidDockerComposeAction._has_comment(
            ('manual-fixuid-install',
             'fixuid-manual-install',
             'custom-fixuid-install',
             'fixuid-custom-install'), lines)

    @staticmethod
    def _has_comment(keywords: Iterable[str], lines: Iterable[str]):
        patterns = [r'^#\s*' + keyword + r'\s*$' for keyword in keywords]
        for line in lines:
            for pattern in patterns:
                if re.match(pattern, line):
                    return True
        return False

    def _apply_fixuid_from_parser(self, parser: CustomDockerfileParser, service: BuildServiceDef):
        if self._has_fixuid_disabled_comment(parser.lines):
            return False

        target = copy_from_url(config.data["fixuid.url"],
                               service.context,
                               "fixuid.tar.gz")
        if target:
            events.file.generated(source=None, target=target)

        manual_entrypoint = self._has_fixuid_manual_entrypoint_comment(parser.lines)
        manual_install = self._has_fixuid_manual_install_comment(parser.lines)
        manual = self._has_fixuid_manual_comment(parser.lines)

        ret = False
        if not manual:
            if not manual_entrypoint:
                cmd, entrypoint = FixuidDockerComposeAction._get_cmd_and_entrypoint(parser)
                fixuid_entrypoint = FixuidDockerComposeAction._add_fixuid_entrypoint(entrypoint)
                if fixuid_entrypoint:
                    parser.entrypoint = fixuid_entrypoint
                if cmd:
                    parser.cmd = cmd
                ret = True

            if not manual_install and self._dockerfile_lines[0] + "\n" not in parser.lines:
                last_instruction_user = parser \
                    .get_last_instruction("USER",
                                          lambda instruction: instruction.get('value') and
                                                              instruction.get('value').lower() != 'root')
                last_instruction_entrypoint = parser.get_last_instruction("ENTRYPOINT")
                if last_instruction_user:
                    parser.add_lines_at(last_instruction_user, *self._dockerfile_lines)
                elif last_instruction_entrypoint:
                    parser.add_lines_at(last_instruction_entrypoint, *self._dockerfile_lines)
                else:
                    parser.add_lines(*self._dockerfile_lines)
                ret = True

        return ret

    def _remove_fixuid_from_parser(self, parser: CustomDockerfileParser, service: BuildServiceDef):
        baseimage_config = FixuidDockerComposeAction._get_image_config(parser.baseimage)
        image_entrypoint = None
        if baseimage_config and 'Entrypoint' in baseimage_config:
            image_entrypoint = json.dumps(baseimage_config['Entrypoint'])

        image_cmd = None
        if baseimage_config and 'Cmd' in baseimage_config:
            image_cmd = json.dumps(baseimage_config['Cmd'])

        if parser.entrypoint:
            parser.entrypoint = FixuidDockerComposeAction._remove_fixuid_entrypoint(parser.entrypoint)

        if parser.entrypoint == image_entrypoint:
            parser.entrypoint = None

        if parser.cmd == image_cmd:
            parser.cmd = None

        removed = False

        fixuid_targz = os.path.join(service.context, "fixuid.tar.gz")
        if os.path.exists(fixuid_targz):
            force_remove(fixuid_targz)
            removed = True

        lines = list(parser.lines)
        for dockerfile_line in self._dockerfile_lines:
            if dockerfile_line + "\n" in lines:
                lines.remove(dockerfile_line + "\n")
                removed = True

        if removed:
            parser.lines = lines

        return removed

    def execute(self, docker_compose_config: dict):
        """
        Execute action
        """
        self.docker_compose_config = docker_compose_config

        for service in self.get_fixuid_services():
            self.apply_fixuid(service)

    def find_fixuid_service(self, dockerfile_filepath: str, include_missing_fixuid=False):
        """
        Find related fixuid service from dockerfile filepath
        """
        for service in self.get_fixuid_services(include_missing_fixuid):
            if os.path.abspath(os.path.join(service.context, service.dockerfile)) == \
                    os.path.abspath(dockerfile_filepath):
                return service
        return None

    def get_fixuid_services(self, include_missing_fixuid=False) -> Iterable[BuildServiceDef]:
        """
        Services where fixuid.tar.gz is available in build context.
        """
        if "services" not in self.docker_compose_config:
            return

        for _, service in self.docker_compose_config.get("services").items():
            if "build" not in service.keys():
                continue

            if isinstance(service["build"], dict):
                build_context = Dotty(service).get("build.context")
            elif isinstance(service["build"], str):
                build_context = service["build"]
            else:
                continue

            if not include_missing_fixuid and not os.path.exists(os.path.join(build_context, "fixuid.yml")):
                continue

            dockerfile = Dotty(service).get("build.dockerfile", "Dockerfile")
            yield BuildServiceDef(build_context, dockerfile)

    @staticmethod
    def _parse_entrypoint(entrypoint):
        as_list = False
        start_quote = ""
        end_quote = ""

        if not entrypoint or entrypoint == "null":
            as_list = True
            entrypoint_list = []
        elif entrypoint.startswith("["):
            as_list = True
            entrypoint_list = json.loads(entrypoint)
        else:
            entrypoint_match = re.compile(r"^(['\"]?)(.*?)(['\"]?)$").match(entrypoint)
            start_quote = entrypoint_match.group(1)
            end_quote = entrypoint_match.group(3)
            entrypoint = entrypoint_match.group(2)
            entrypoint_list = shlex.split(entrypoint)

        return entrypoint_list, as_list, start_quote, end_quote

    @staticmethod
    def _add_fixuid_entrypoint(entrypoint):
        entrypoint_list, as_list, start_quote, end_quote = \
            FixuidDockerComposeAction._parse_entrypoint(entrypoint)

        fixuid_entrypoint_prefix = ["fixuid", "-q"]
        if entrypoint_list[:len(fixuid_entrypoint_prefix)] != fixuid_entrypoint_prefix:
            entrypoint_list = ["fixuid", "-q"] + entrypoint_list
        if as_list:
            entrypoint = json.dumps(entrypoint_list)
        else:
            entrypoint = " ".join(entrypoint_list)
            entrypoint = "%s%s%s" % (start_quote, entrypoint, end_quote)

        return entrypoint

    @staticmethod
    def _remove_fixuid_entrypoint(entrypoint):
        entrypoint_list, as_list, start_quote, end_quote = \
            FixuidDockerComposeAction._parse_entrypoint(entrypoint)

        fixuid_entrypoint_prefix = ["fixuid", "-q"]
        if entrypoint_list[:len(fixuid_entrypoint_prefix)] == fixuid_entrypoint_prefix:
            entrypoint_list = entrypoint_list[len(fixuid_entrypoint_prefix):]
        if as_list:
            entrypoint = json.dumps(entrypoint_list)
        else:
            entrypoint = " ".join(entrypoint_list)
            entrypoint = "%s%s%s" % (start_quote, entrypoint, end_quote)

        return entrypoint



# -*- coding: utf-8 -*-
from typing import ClassVar, Iterable

from ddb.action import Action
from ddb.feature import Feature
from .actions import FixFilePermissionsAction
from .schema import GitSchema


class GitFeature(Feature):
    """
    Git support.
    """

    @property
    def name(self) -> str:
        return "git"

    @property
    def dependencies(self) -> Iterable[str]:
        return ["core"]

    @property
    def schema(self) -> ClassVar[GitSchema]:
        return GitSchema

    @property
    def actions(self) -> Iterable[Action]:
        return (
            FixFilePermissionsAction(),
        )



# -*- coding: utf-8 -*-
import os
import patch
from typing import Union

from cookiecutter.config import DEFAULT_CONFIG
from cookiecutter.main import cookiecutter

from ddb.action import Action
from ddb.config import config
from ddb.context import context
from ddb.event import events


class CookiecutterAction(Action):
    """
    Download a cookiecutter template.
    """

    @property
    def name(self) -> str:
        return "cookiecutter:download"

    @property
    def event_bindings(self):
        return events.phase.download

    def execute(self):
        """
        Execute action
        """
        config_parameters = 'cookiecutters_dir', 'replay_dir', 'default_context'

        defaults = {}
        for parameter in config_parameters:
            defaults[parameter] = DEFAULT_CONFIG[parameter]

        cookiecutter_config = config.data.get('cookiecutter')
        try:
            if cookiecutter_config:
                for parameter in config_parameters:
                    try:
                        DEFAULT_CONFIG[parameter] = cookiecutter_config[parameter]
                    except KeyError:
                        pass

            templates = cookiecutter_config.get('templates')
            if templates:
                for template in templates:
                    self._generate_template(template)
        finally:
            for parameter in config_parameters:
                DEFAULT_CONFIG[parameter] = defaults[parameter]

    @staticmethod
    def _generate_template(template):
        if template['no_input'] is None:
            template['no_input'] = True

        kwargs = {k: v for k, v in template.items() if v is not None}

        ret = cookiecutter(**kwargs)
        output_dir = os.path.relpath(ret, '.')

        context.log.success(f"{template['template']} -> {output_dir}")

        for patch_file in os.listdir(output_dir):
            if patch_file.endswith('.patch'):
                patch_file = os.path.join(output_dir, patch_file)
                if os.path.isfile(patch_file):
                    context.log.notice("Applying patch %s in %s", patch_file, output_dir)
                    try:
                        pset = patch.fromfile(patch_file)
                        pset.apply(root=output_dir)
                        context.log.success("Patch %s applied in %s", patch_file, output_dir)
                    except Exception as exc:  # pylint:disable=broad-except
                        context.log.error("Patch %s has failed to be applied in %s: %s", patch_file, output_dir, exc)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import io
import os
import re
import sys

from setuptools import setup, find_packages

try:
    from semantic_release import setup_hook

    setup_hook(sys.argv)
except ImportError:
    pass

here = os.path.abspath(os.path.dirname(__file__))

with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    readme = f.read()

with io.open(os.path.join(here, 'CHANGELOG.md'), encoding='utf-8') as f:
    history = f.read()

project_dir = os.path.dirname(os.path.realpath(__file__))

install_requires = [
    "colorama",
    "colorlog",
    "verboselogs",
    "netifaces",
    "requests",
    "PyYAML",
    "marshmallow",
    "dotty-dict",
    "deepmerge",
    "diskcache",
    "python-slugify",
    "dictdiffer",
    "Jinja2",
    "braceexpand",
    "toposort",
    "cookiecutter",
    "jsonnet-binary",
    "zgitignore",
    "cached-property",
    "dockerfile-parse",
    "cfssl>=0.0.3b243",
    "cryptography",
    "watchdog",
    "gitpython",
    "chmod-monkey>=1.1.1",
    "paramiko",
    "wcmatch",
    "marshmallow-union",
    "progress",
    "ordered-set",
    "patch",
    "semver",
    "distro"
]

dev_requires = [
    "python-semantic-release",
    "pyinstaller",
    "commitizen",
    "pre-commit",
    "tox",
    "astroid",
    "pylint",
    "docker[tls]",
    "docker-compose",
    "pytest",
    "coverage",
    "waiting",
    "pytest-mock",
    "pytest-cov",
    "pex"
]

package_data = []

entry_points = {
    'console_scripts': [
        'ddb = ddb.__main__:console_script'
    ],
    "pytest11": [
        "docker_compose=pytest_docker_compose:plugin",
    ],
}

with io.open('ddb/__version__.py', 'r') as f:
    version = re.search(r'^__version__\st*=\s*[\'"]([^\'"]*)[\'"]$', f.read(), re.MULTILINE).group(1)

args = dict(name='docker-devbox-ddb',
            version=version,
            description='ddb - Erase environment differences, make developers happy !',
            long_description=readme + '\n' * 2 + history,
            long_description_content_type='text/markdown',
            # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
            classifiers=['Development Status :: 5 - Production/Stable',
                         'License :: OSI Approved :: MIT License',
                         'Operating System :: OS Independent',
                         'Intended Audience :: Developers',
                         'Programming Language :: Python :: 3',
                         'Programming Language :: Python :: 3.6',
                         'Programming Language :: Python :: 3.7',
                         'Programming Language :: Python :: 3.8',
                         'Programming Language :: Python :: 3.9',
                         'Topic :: Software Development :: Libraries :: Python Modules'
                         ],
            keywords='docker docker-compose development environment devops templates jsonnet jinja watch',
            author='Rémi Alvergnat',
            author_email='remi.alvergnat@gfi.world',
            url='https://github.com/inetum-orleans/docker-devbox-ddb',
            download_url='https://pypi.python.org/packages/source/g/docker-devbox-ddb/docker-devbox-ddb-%s.tar.gz' % version,
            license='MIT',
            packages=find_packages(),
            include_package_data=True,
            install_requires=install_requires,
            entry_points=entry_points,
            zip_safe=True,
            extras_require={
                'dev': dev_requires
            })

setup(**args)



# -*- coding: utf-8 -*-
import os

import cfssl as cfssl_client

from .cfssl import checksums, writer
from ...action import Action
from ...config import config
from ...context import context
from ...event import events
from ...utils.file import force_remove, copy_if_different


class GenerateCertAction(Action):
    """
    Generate a certificate for a domain.
    """

    @property
    def event_bindings(self):
        return events.certs.generate

    @property
    def name(self) -> str:
        return "certs:generate"

    @staticmethod
    def execute(domain, wildcard=True):
        """
        Generate a certificate
        """
        if config.data.get('certs.type') == 'cfssl':
            GenerateCertAction.generate_cfssl(domain, wildcard)

    @staticmethod
    def generate_cfssl(domain, wildcard=True):
        """
        Generate domain certificate with CFSSL.
        """
        certificate_path, private_key_path = writer.get_certs_path(domain,
                                                                   config.data['certs.cfssl.writer'],
                                                                   config.data['certs.destination'])
        if not os.path.exists(certificate_path) or not os.path.exists(private_key_path):
            client_config = config.data.get('certs.cfssl.server')

            client = cfssl_client.CFSSL(**client_config)

            certificate_request = cfssl_client.CertificateRequest()
            certificate_request.common_name = domain
            certificate_request.hosts = (domain, "*.%s" % domain) if wildcard else domain

            response = client.new_cert(certificate_request)
            checksums.validate_checksums(response)

            destination = config.data['certs.destination']
            if not os.path.isdir(destination):
                os.makedirs(destination)

            generated = writer.write_files(response,
                                           domain,
                                           None,
                                           None,
                                           config.data['certs.cfssl.writer'],
                                           config.data['certs.destination'],
                                           config.data['certs.cfssl.append_ca_certificate'],
                                           client,
                                           config.data['certs.cfssl.verify_checksum'])

            context.log.success("TLS certificate generated for domain %s", domain)

            for generated_file in generated.values():
                events.file.generated(source=None, target=generated_file)

            private_key_path, certificate_path = generated['private_key'], generated['certificate']
            events.certs.generated(domain=domain, wildcard=wildcard, private_key=certificate_path,
                                   certificate=private_key_path)
        else:
            context.log.notice("TLS certificate exists for domain %s", domain)

        events.certs.available(domain=domain, wildcard=wildcard, private_key=private_key_path,
                               certificate=certificate_path)


class RemoveCertAction(Action):
    """
    Remove a certificate for a domain.
    """

    @property
    def event_bindings(self):
        return events.certs.remove

    @property
    def name(self) -> str:
        return "certs:remove"

    @staticmethod
    def execute(domain):
        """
        Remove certificate
        """
        if RemoveCertAction.remove_cfssl(domain):
            events.certs.removed(domain=domain)

    @staticmethod
    def remove_cfssl(domain):
        """
        Remove domain certificate generated with CFSSL.
        """
        certificate_path, private_key_path = writer.get_certs_path(domain,
                                                                   config.data['certs.cfssl.writer'],
                                                                   config.data['certs.destination'])

        removed = False
        if os.path.exists(certificate_path):
            force_remove(certificate_path)
            events.file.deleted(certificate_path)
            removed = True

        if os.path.exists(private_key_path):
            force_remove(private_key_path)
            events.file.deleted(certificate_path)
            removed = True

        return removed


class GenerateSignerCertificateAction(Action):
    """
    Generate the signer certificate, used to sign domains certificates.
    """

    @property
    def event_bindings(self):
        return events.certs.generated

    @property
    def name(self) -> str:
        return "certs:generate-signer"

    @property
    def disabled(self) -> bool:
        return config.data.get('certs.type') != 'cfssl'

    @staticmethod
    def execute(*args, **kwargs):
        """
        Generate certificate with CFSSL.
        """
        output = ".signer"
        certificate_path, _ = writer.get_certs_path(output,
                                                    config.data['certs.cfssl.writer'],
                                                    config.data['certs.destination'])

        if not os.path.exists(certificate_path):
            client_config = config.data.get('certs.cfssl.server')

            client = cfssl_client.CFSSL(**client_config)
            response = client.info('')

            destination = config.data['certs.destination']
            if not os.path.isdir(destination):
                os.makedirs(destination)

            generated = writer.write_files(response,
                                           output,
                                           None,
                                           None,
                                           config.data['certs.cfssl.writer'],
                                           config.data['certs.destination'],
                                           False,
                                           client,
                                           False)

            context.log.success("Signer TLS certificate generated")

            for generated_file in generated.values():
                events.file.generated(source=None, target=generated_file)
        else:
            context.log.notice("Signer TLS certificate exists")

        signer_destinations = config.data['certs.signer_destinations']
        if signer_destinations:
            certificate = os.path.join(config.data['certs.destination'],
                                       output + config.data['certs.cfssl.writer']['filenames']['certificate'] % '')

            for signer_destination in signer_destinations:
                dirname = os.path.dirname(signer_destination)
                if not os.path.isdir(dirname):
                    os.makedirs(dirname)
                if copy_if_different(certificate, signer_destination) or config.eject:
                    events.file.generated(source=certificate, target=signer_destination)



import os
from shutil import copy

import pytest
from _pytest.capture import CaptureFixture

from ddb.__main__ import main, reset
from ddb.config import Config, config

docker_compose_bin = "docker-compose" if os.name != "nt" else "docker-compose.exe"


class TestBinaries:
    def test_docker_binaries(self, project_loader, capsys: CaptureFixture):
        project_loader("docker")

        exceptions = main(["configure"])
        assert not exceptions

        exceptions = main(["run", "psql"])
        assert not exceptions

        assert os.path.exists(os.path.join(os.getcwd(), '.bin', 'psql'))

        output = capsys.readouterr()
        assert output.out.strip() == docker_compose_bin + " run --rm --workdir=/workdir/. db psql"

    def test_docker_binaries_with_custom_options(self, project_loader, capsys: CaptureFixture):
        project_loader("docker")

        exceptions = main(["configure"])
        assert not exceptions

        os.environ['DDB_RUN_OPTS'] = "-e TEST"

        exceptions = main(["run", "psql"])
        assert not exceptions

        assert os.path.exists(os.path.join(os.getcwd(), '.bin', 'psql'))

        output = capsys.readouterr()
        assert output.out.strip() == docker_compose_bin + " run --rm --workdir=/workdir/. -e TEST db psql"

    def test_docker_binaries_exe(self, project_loader, capsys: CaptureFixture):
        project_loader("docker_exe")

        exceptions = main(["configure"])
        assert not exceptions

        exceptions = main(["run", "psql"])
        assert not exceptions

        assert os.path.exists(os.path.join(os.getcwd(), '.bin', 'psql'))

        output = capsys.readouterr()
        assert output.out.strip() == docker_compose_bin + " exec --workdir=/workdir/. db psql"

    def test_docker_binaries_entrypoint(self, project_loader, capsys: CaptureFixture):
        project_loader("docker_entrypoint")

        exceptions = main(["configure"])
        assert not exceptions

        exceptions = main(["run", "psql"])
        assert not exceptions

        assert os.path.exists(os.path.join(os.getcwd(), '.bin', 'psql'))

        output = capsys.readouterr()
        assert output.out.strip() == docker_compose_bin + " run --rm --workdir=/workdir/. --entrypoint=/custom/ep db psql"

    def test_docker_binaries_global(self, project_loader, capsys: CaptureFixture):
        project_loader("docker_global")

        exceptions = main(["configure"])
        assert not exceptions

        exceptions = main(["run", "psql"])
        assert not exceptions

        assert not os.path.exists(os.path.join(os.getcwd(), '.bin', 'psql'))
        assert os.path.exists(os.path.join(os.getcwd(), '..', 'home', '.bin', 'psql'))

        output = capsys.readouterr()
        assert output.out.strip() == docker_compose_bin + " run --rm --workdir=/workdir/. db psql"

    def test_docker_binaries_global_default(self, project_loader, capsys: CaptureFixture):
        project_loader("docker_global_default")

        exceptions = main(["configure"])
        assert not exceptions

        exceptions = main(["run", "psql"])
        assert not exceptions

        assert not os.path.exists(os.path.join(os.getcwd(), '.bin', 'psql'))
        assert os.path.exists(os.path.join(os.getcwd(), '..', 'home', '.bin', 'psql'))

        output = capsys.readouterr()
        assert output.out.strip() == docker_compose_bin + " run --rm --workdir=/workdir/. db psql"

    def test_docker_binaries_with_clear_cache(self, project_loader, capsys: CaptureFixture):
        project_loader("docker")

        main(["--clear-cache", "configure"])

        exceptions = main(["run", "psql"])
        assert not exceptions

        assert os.path.exists(os.path.join(os.getcwd(), '.bin', 'psql'))

    def test_docker_binaries_removed(self, project_loader, capsys: CaptureFixture):
        Config.defaults = None

        project_loader("docker_removal")

        main(["configure"])

        if os.name == 'nt':
            assert os.path.isfile(os.path.join(".bin", "psql.bat"))
        else:
            assert os.path.isfile(os.path.join(".bin", "psql"))

        exceptions = main(["run", "psql"])
        assert not exceptions

        assert os.path.exists(os.path.join(os.getcwd(), '.bin', 'psql'))

        copy(os.path.join("..", "replacement", "docker-compose.yml.jsonnet"), 'docker-compose.yml.jsonnet')

        main(["configure"])

        assert not os.path.isfile(os.path.join(".bin", "psql"))

        exceptions = main(["run", "psql"])
        assert exceptions

        assert isinstance(exceptions[0], ValueError)
        assert str(exceptions[0]) == 'Binary name "psql" is not registered'

    @pytest.mark.parametrize(
        'relative_cwd,expected',
        [('node10', 'node10'), ('node12', 'node12'), (None, 'node14')]
    )
    def test_docker_binaries_condition(self, relative_cwd, expected, project_loader, capsys: CaptureFixture):
        project_loader("docker_condition")

        cwd = os.getcwd()
        if relative_cwd:
            cwd = os.path.join(cwd, relative_cwd)
            os.chdir(cwd)
            assert os.getcwd() == cwd

        try:
            main(["--clear-cache", "configure"], reset_disabled=True)
            assert config.cwd == cwd
        finally:
            reset()

        exceptions = main(["run", "node", "--version"])
        assert not exceptions

        outerr = capsys.readouterr()
        assert expected in outerr.out



# -*- coding: utf-8 -*-
from typing import Callable, Dict, List, Union


class Bus:
    """
    Event bus.

    Can be used to register/unregister events listeners, and emit events.
    """
    _named_listeners = {}  # type: Dict[str, List[Callable]]
    _listeners = []  # type: List[Callable]

    def on(self, event_name: Union[str, None], listener: Callable):  # pylint:disable=invalid-name
        """
        Register a listener on given event name.
        """
        if event_name:
            if event_name not in self._named_listeners:
                self._named_listeners[event_name] = []
            listeners = self._named_listeners[event_name]
            listeners.append(listener)
        else:
            self._listeners.append(listener)
        return lambda: self.off(event_name, listener)

    def off(self, event_name: Union[str, None], listener: Callable):
        """
        Unregister a listener from given event name.
        """
        if event_name:
            listeners = self._named_listeners.get(event_name)
            if listeners:
                listeners.remove(listener)
        else:
            self._listeners.remove(listener)

    def clear(self):
        """
        Unregister all listeners.
        """
        self._named_listeners.clear()
        self._listeners.clear()

    def emit(self, event_name: str, *args, **kwargs):
        """
        Emit a new event with given payload.
        """
        named_listeners = self._named_listeners.get(event_name)
        if named_listeners:
            for listener in named_listeners:
                listener(*args, **kwargs)

        listeners = self._listeners
        for listener in listeners:
            listener(event_name, *args, **kwargs)

    def has_named_listeners(self, event_name: str) -> bool:
        """
        Check if event_name has at least one named listener registered
        """
        named_listeners = self._named_listeners.get(event_name)
        return bool(named_listeners)



import os

from ddb.__main__ import load_registered_features, main
from ddb.feature import features
from ddb.feature.core import CoreFeature
from ddb.feature.gitignore import UpdateGitignoreAction, GitignoreFeature
from tests.utilstest import expect_gitignore


class TestUpdateGitIgnoreAction:
    def test_get_relative_path(self, project_loader):
        assert UpdateGitignoreAction._get_relative_path(target='/home/vagrant/sub/to-ignore.yml',
                                                        gitignore='/home/vagrant/.gitignore') == '/sub/to-ignore.yml'
        assert UpdateGitignoreAction._get_relative_path(target='/home/vagrant/to-ignore.yml',
                                                        gitignore='/home/vagrant/.gitignore') == '/to-ignore.yml'
        assert UpdateGitignoreAction._get_relative_path(target='/home/vagrant/to-ignore.yml',
                                                        gitignore='/home/vagrant/.gitignore',
                                                        first_slash=True) == '/to-ignore.yml'
        assert UpdateGitignoreAction._get_relative_path(target='/home/vagrant/to-ignore.yml',
                                                        gitignore='/home/vagrant/.gitignore',
                                                        first_slash=False) == 'to-ignore.yml'

    def test_empty_project_without_core(self, project_loader):
        project_loader("empty")

        features.register(GitignoreFeature())
        load_registered_features()

        action = UpdateGitignoreAction()
        action.execute(target="./to-ignore.yml")
        action.execute(target="./to-ignore-2.yml")

        assert os.path.exists('.gitignore')
        assert expect_gitignore('.gitignore', '/to-ignore.yml', '/to-ignore-2.yml')

    def test_empty_project_with_core(self, project_loader):
        project_loader("empty")

        features.register(CoreFeature())
        features.register(GitignoreFeature())
        load_registered_features()

        action = UpdateGitignoreAction()
        action.execute(target="./to-ignore.yml")

        assert os.path.exists('.gitignore')
        assert expect_gitignore('.gitignore', '/to-ignore.yml')

        action.remove("./to-ignore.yml")

        assert not os.path.exists('.gitignore')

    def test_empty_skip_outside_additions(self, project_loader):
        project_loader("empty")

        features.register(CoreFeature())
        features.register(GitignoreFeature())
        load_registered_features()

        action = UpdateGitignoreAction()
        action.execute(target="../outside/to-ignore.yml")

        assert not os.path.exists('.gitignore')

    def test_already_ignored(self, project_loader):
        project_loader("already_ignored")

        assert os.path.exists('.gitignore')
        with open('.gitignore', 'r') as f:
            expected_gitignore = f.read()

        features.register(GitignoreFeature())
        load_registered_features()

        action = UpdateGitignoreAction()
        action.execute(target="./exists.yml")

        assert os.path.exists('.gitignore')
        with open('.gitignore', 'r') as f:
            gitignore = f.read()

        assert gitignore == expected_gitignore

    def test_remove_empty_block(self, project_loader):
        project_loader("already_ignored")

        assert os.path.exists('.gitignore')
        with open('.gitignore', 'r') as f:
            expected_gitignore = f.read()

        features.register(GitignoreFeature())
        load_registered_features()

        action = UpdateGitignoreAction()
        action.execute(target="./another.yml")

        action.remove("./another.yml")
        assert os.path.exists('.gitignore')
        with open('.gitignore', 'r') as f:
            gitignore = f.read()

        assert gitignore == expected_gitignore

    def test_keep_negated(self, project_loader):
        project_loader("keep_negated")

        assert os.path.exists('.gitignore')
        with open('.gitignore', 'r') as f:
            expected_gitignore = f.read()

        features.register(GitignoreFeature())
        load_registered_features()

        action = UpdateGitignoreAction()
        action.execute(target="./exists.yml")

        assert os.path.exists('.gitignore')
        with open('.gitignore', 'r') as f:
            gitignore = f.read()

        assert gitignore == expected_gitignore

    def test_templates(self, project_loader):
        project_loader("empty")

        main(["configure"])

        assert os.path.exists(os.path.join('.gitignore'))
        assert expect_gitignore('.gitignore', '/*ddb.local.*')

        # TODO faire fonctionner ce test
        # config.data.__setitem__('gitignore.enforce', ['config.txt'])
        # main(["configure"])
        #
        # assert os.path.exists(os.path.join('.gitignore'))
        # assert not expect_gitignore('.gitignore', '*ddb.local.*')
        # assert expect_gitignore('.gitignore', 'config.txt')

    def test_stage_env(self, project_loader):
        project_loader("stage-env")

        gitignore_feature = GitignoreFeature()

        features.register(CoreFeature())
        features.register(gitignore_feature)
        load_registered_features()

        assert gitignore_feature.disabled



# -*- coding: utf-8 -*-
from functools import wraps

from .bus import Bus

bus = Bus()


def event(event_name):
    """
    A decorator for event definition
    :param event_name: name of the event
    """

    def decorator(function):
        function.name = event_name

        @wraps(function)
        def perform(*args, **kwargs):
            bus.emit(event_name, *args[1:], **kwargs)

        return perform

    return decorator



import os
import posixpath
import shlex
from pathlib import Path
from typing import Optional, Iterable

from ddb.utils.simpleeval import simple_eval

from ddb.binary.binary import AbstractBinary
from ddb.config import config
from ddb.feature.docker.utils import get_mapped_path
from ddb.utils.docker import DockerUtils
from ddb.utils.process import effective_command


class DockerBinary(AbstractBinary):  # pylint:disable=too-many-instance-attributes
    """
    Binary to run docker compose command.
    """

    def __init__(self,
                 name: str,
                 docker_compose_service: str,
                 workdir: Optional[str] = None,
                 options: Optional[str] = None,
                 options_condition: Optional[str] = None,
                 condition: Optional[str] = None,
                 args: Optional[str] = None,
                 exe: bool = False,
                 entrypoint: Optional[str] = None,
                 global_: bool = False):
        super().__init__(name)
        self.docker_compose_service = docker_compose_service
        self.workdir = workdir
        self.options = options
        self.options_condition = options_condition
        self.condition = condition
        self.args = args
        self.exe = exe
        self.entrypoint = entrypoint
        self._global = global_

    @staticmethod
    def simple_eval_options(*args):
        """
        Retrieve the simple_eval options
        """
        return dict(functions={},
                    names={"args": " ".join(args),
                           "argv": args,
                           "config": config,
                           "cwd": str(Path(config.cwd).as_posix()) if config.cwd else None,
                           "project_cwd": str(Path(config.project_cwd).as_posix()) if config.project_cwd else None})

    def command(self, *args) -> Iterable[str]:
        cwd = config.cwd if config.cwd else os.getcwd()
        real_cwd = os.path.realpath(cwd)
        real_project_home = os.path.realpath(config.paths.project_home)

        if real_cwd.startswith(real_project_home):
            params = ["exec"] if hasattr(self, "exe") and self.exe else ["run", "--rm"]

            if self.workdir:
                relpath = os.path.relpath(cwd, config.paths.project_home)
                container_workdir = posixpath.join(self.workdir, relpath)
                params.append(f"--workdir={container_workdir}")
        else:
            # cwd is outside of project home.
            project_relpath = os.path.relpath(config.paths.project_home, config.cwd if config.cwd else os.getcwd())
            params = ["-f", os.path.join(project_relpath, "docker-compose.yml"), "run", "--rm"]
            if self.workdir:
                mapped_cwd = get_mapped_path(real_cwd)
                params.append(f"--volume={mapped_cwd}:{self.workdir}")
                params.append(f"--workdir={self.workdir}")

        if self.entrypoint:
            params.append(f"--entrypoint={self.entrypoint}")

        if os.environ.get('DDB_RUN_OPTS'):
            params.extend([shlex.quote(param) for param in shlex.split(os.environ.get('DDB_RUN_OPTS'))])

        self.add_options_to_params(params, self.options, self.options_condition, *args)

        params.append(self.docker_compose_service)
        if self.args:
            params.extend([shlex.quote(param) for param in shlex.split(self.args)])

        command = effective_command("docker-compose", *params)
        return command

    def add_options_to_params(self, params, options, condition, *args):
        """
        Add options to params if condition is fulfilled
        :param params: the list of parameters
        :param options: options to add
        :param condition: the condition to fulfilled
        :param args: the list of args of binary call
        :return:
        """
        if condition is not None and options is not None:
            if simple_eval(condition, **self.simple_eval_options(*args)):
                params.extend(shlex.split(options))
        else:
            if options is not None:
                params.extend(shlex.split(options))

    def before_run(self, *args):
        if self.exe and not DockerUtils.is_container_up(self.docker_compose_service):
            DockerUtils.service_up(self.docker_compose_service)

    def should_run(self, *args) -> bool:
        if self.condition:
            return bool(simple_eval(self.condition, **self.simple_eval_options(*args)))
        return super().should_run(*args)

    @property
    def global_(self) -> bool:
        return self._global

    def __lt__(self, other):
        """
        This is used to order binaries in run feature action.
        """
        if not isinstance(other, DockerBinary):
            return True
        return self.condition and not other.condition

    def __gt__(self, other):
        """
        This is used to order binaries in run feature action.
        """
        if not isinstance(other, DockerBinary):
            return False
        return not self.condition and other.condition

    def __eq__(self, other):  # pylint:disable=too-many-return-statements
        """
        Check if given binary is the same as the current one
        :param binary:
        :return: True or False depending on it's the same or not
        """
        if not super().__eq__(other):
            return False
        if not isinstance(other, DockerBinary):
            return False
        if self.docker_compose_service != other.docker_compose_service:
            return False
        if self.workdir != other.workdir:
            return False
        if self.options != other.options:
            return False
        if self.options_condition != other.options_condition:
            return False
        if self.condition != other.condition:
            return False
        if self.args != other.args:
            return False
        if self.exe != other.exe:
            return False
        if self.entrypoint != other.entrypoint:
            return False

        return True

    def __hash__(self):
        return hash((super().__hash__(),
                     self.docker_compose_service,
                     self.workdir,
                     self.options,
                     self.options_condition,
                     self.condition,
                     self.args,
                     self.exe,
                     self.entrypoint))



import os
import pytest

from ddb.__main__ import load_registered_features
from ddb.feature import features
from ddb.feature.certs import CertsFeature, GenerateCertAction
from ddb.feature.certs.actions import RemoveCertAction
from ddb.feature.core import CoreFeature
from tests.utilstest import setup_cfssl


class TestCertsFeature:
    @pytest.mark.docker
    def test_empty_project_without_core(self, project_loader, module_scoped_container_getter):
        project_loader("empty")

        features.register(CertsFeature())
        load_registered_features()

        setup_cfssl(module_scoped_container_getter)

        generate_action = GenerateCertAction()
        generate_action.execute(domain="testing.test")

        assert os.path.exists(os.path.join(".certs", "testing.test.crt"))
        assert os.path.exists(os.path.join(".certs", "testing.test.key"))

        remove_action = RemoveCertAction()
        remove_action.execute(domain="testing.test")
        assert not os.path.exists(os.path.join(".certs", "testing.test.crt"))
        assert not os.path.exists(os.path.join(".certs", "testing.test.key"))

    @pytest.mark.docker
    def test_empty_project_with_core(self, project_loader, module_scoped_container_getter):
        project_loader("empty")

        features.register(CoreFeature())
        features.register(CertsFeature())
        load_registered_features()

        setup_cfssl(module_scoped_container_getter)

        generate_action = GenerateCertAction()
        generate_action.execute(domain="testing.test")

        assert os.path.exists(os.path.join(".certs", "testing.test.crt"))
        assert os.path.exists(os.path.join(".certs", "testing.test.key"))

        with open(os.path.join(".certs", "testing.test.crt")) as crt_file:
            assert 'CERTIFICATE' in crt_file.read()

        with open(os.path.join(".certs", "testing.test.key")) as key_file:
            assert 'PRIVATE KEY' in key_file.read()

        remove_action = RemoveCertAction()
        remove_action.execute(domain="testing.test")
        assert not os.path.exists(os.path.join(".certs", "testing.test.crt"))
        assert not os.path.exists(os.path.join(".certs", "testing.test.key"))

    @pytest.mark.docker
    def test_existing_does_nothing(self, project_loader, module_scoped_container_getter):
        project_loader("existing")

        features.register(CoreFeature())
        features.register(CertsFeature())
        load_registered_features()

        setup_cfssl(module_scoped_container_getter)

        assert os.path.exists(os.path.join(".certs", "testing.test.crt"))
        assert os.path.exists(os.path.join(".certs", "testing.test.key"))

        with open(os.path.join(".certs", "testing.test.crt")) as crt_file:
            crt = crt_file.read()

        with open(os.path.join(".certs", "testing.test.key")) as key_file:
            key = key_file.read()

        generate_action = GenerateCertAction()
        generate_action.execute(domain="testing.test")

        with open(os.path.join(".certs", "testing.test.crt")) as crt_file:
            assert crt == crt_file.read()

        with open(os.path.join(".certs", "testing.test.key")) as key_file:
            assert key == key_file.read()

        remove_action = RemoveCertAction()
        remove_action.execute(domain="testing.test")

        assert not os.path.exists(os.path.join(".certs", "testing.test.crt"))
        assert not os.path.exists(os.path.join(".certs", "testing.test.key"))



# -*- coding: utf-8 -*-
from marshmallow import Schema, fields


class DisableableSchema(Schema):
    """
    Base feature schema.
    """
    disabled = fields.Bool(default=False)


class FeatureSchema(DisableableSchema):
    """
    Base feature schema.
    """



# -*- coding: utf-8 -*-
from typing import Iterable, ClassVar

from .actions import FixuidDockerComposeAction
from .schema import FixuidSchema
from ..feature import Feature
from ..schema import FeatureSchema
from ...action import Action


class FixuidFeature(Feature):
    """
    Auto-configure fixuid from docker-compose.yml services (https://github.com/boxboat/fixuid)
    """

    @property
    def name(self) -> str:
        return "fixuid"

    @property
    def dependencies(self) -> Iterable[str]:
        return ["core", "docker[optional]"]

    @property
    def schema(self) -> ClassVar[FeatureSchema]:
        return FixuidSchema

    @property
    def actions(self) -> Iterable[Action]:
        return (
            FixuidDockerComposeAction(),
        )



# -*- coding: utf-8 -*-
from typing import Iterable, ClassVar

from .actions import CopyAction
from .schema import CopySchema
from ..feature import Feature
from ..schema import FeatureSchema
from ...action import Action


class CopyFeature(Feature):
    """
    Copy files from local filesystem or remote URL to one or many directories.
    """

    @property
    def name(self) -> str:
        return "copy"

    @property
    def dependencies(self) -> Iterable[str]:
        return ["core"]

    @property
    def schema(self) -> ClassVar[FeatureSchema]:
        return CopySchema

    @property
    def actions(self) -> Iterable[Action]:
        return (
            CopyAction(),
        )



# -*- coding: utf-8 -*-
from marshmallow import fields, Schema

from ddb.feature.schema import FeatureSchema


class UserSchema(Schema):
    """
    User schema
    """
    uid = fields.Integer(required=True, default=None)  # default is set in feature _configure_defaults
    gid = fields.Integer(required=True, default=None)  # default is set in feature _configure_defaults
    name = fields.String(required=False, allow_none=True, default=None)  # default is set in feature _configure_defaults
    group = fields.String(required=False, allow_none=True,
                          default=None)  # default is set in feature _configure_defaults


class DockerSchema(FeatureSchema):
    """
    Docker schema.
    """
    ip = fields.String(required=True, default=None)  # default is set in feature _configure_defaults
    interface = fields.String(required=True, default="docker0")
    user = fields.Nested(UserSchema(), default=UserSchema())
    path_mapping = fields.Dict(required=False)  # default is set in feature _configure_defaults



# https://github.com/pytest-docker-compose/pytest-docker-compose/pull/64
# pylint: skip-file

import os
import time
from typing import List
from pathlib import Path
import warnings
from datetime import datetime

import pytest
from compose.cli.command import project_from_options
from compose.config.errors import ComposeFileNotFound
from compose.container import Container
from compose.project import Project
from compose.service import ImageType

class ContainersAlreadyExist(Exception):
    """Raised when running containers are unexpectedly found"""
    pass


class ContainerDoesntExist(Exception):
    """Raised when container doesn't exists"""
    pass


class ContainerNotRunning(Exception):
    """Raised when container doesn't exists"""
    pass


__all__ = [
    "DockerComposePlugin",
    "NetworkInfo",
    "plugin",
]


class NetworkInfo:
    def __init__(self, container_port: str, hostname: str, host_port: int, ):
        """
        Container for info about how to connect to a service exposed by a
        Docker container.

        :param container_port: Port (and usually also protocol name) exposed
        internally on the container.
        :param hostname: Hostname to use when accessing this service.
        :param host_port: Port number to use when accessing this service.
        """
        self.container_port = container_port
        self.hostname = hostname
        self.host_port = host_port


def create_network_info_for_container(container: Container):
    """
    Generates :py:class:`NetworkInfo` instances corresponding to all available
    port bindings in a container
    """
    # If ports are exposed by the docker container but not by docker expose
    # container.ports looks like this:
    # container.ports == {'4369/tcp': None,
    # '5984/tcp': [{'HostIp': '0.0.0.0', 'HostPort': '32872'}],
    # '9100/tcp': None}
    return [NetworkInfo(container_port=container_port,
                        hostname=port_config["HostIp"] or "localhost",
                        host_port=port_config["HostPort"], )
            for container_port, port_configs in container.ports.items()
            if port_configs is not None for port_config in port_configs]


class DockerComposePlugin:
    """
    Integrates docker-compose into pytest integration tests.
    """

    def __init__(self):
        self.function_scoped_container_getter = self.generate_scoped_containers_fixture('function')
        self.class_scoped_container_getter = self.generate_scoped_containers_fixture('class')
        self.module_scoped_container_getter = self.generate_scoped_containers_fixture('module')
        self.session_scoped_container_getter = self.generate_scoped_containers_fixture('session')

    # noinspection SpellCheckingInspection
    @staticmethod
    def pytest_addoption(parser):
        """
        Adds custom options to the ``pytest`` command.

        https://docs.pytest.org/en/latest/writing_plugins.html#_pytest.hookspec.pytest_addoption
        """
        group = parser.getgroup("docker_compose", "integration tests")

        group.addoption(
            "--docker-compose",
            action='append',
            dest="docker_compose",
            default=[],
            help="Path to docker-compose.yml file, or directory containing same.",
        )

        group.addoption(
            "--docker-compose-directory",
            dest="project_dir",
            default=[],
            help="Path to docker-compose directory",
        )

        group.addoption("--docker-compose-remove-volumes", action="store_true",
                        default=False, help="Remove docker container volumes after tests")

        group.addoption("--docker-compose-no-build", action="store_true",
                        default=False, help="Boolean to not build docker containers")

        group.addoption("--use-running-containers", action="store_true",
                        default=False, help="Boolean to use a running set of containers "
                                            "instead of calling 'docker-compose up'")

    @classmethod
    def project_from_options_for_each_dir(cls, dirs, *args, **kwargs):
        """
        Create the Docker project from options, trying each project_dirs.
        """
        exc = None
        for project_dir in dirs:
            try:
                project = project_from_options(project_dir=project_dir, *args, **kwargs)
                return project, project_dir
            except ComposeFileNotFound as local_exc:
                exc = local_exc
        raise exc

    @classmethod
    def projectdir_from_basedirs_and_docker_compose_options(cls, basedirs, docker_compose_options):
        """
        Get the project directory from possible basedirs and given --docker-compose options.
        """
        for basedir in basedirs:
            found = True
            for docker_compose_option in docker_compose_options:
                docker_compose_path = Path(basedir).joinpath(docker_compose_option)
                if not docker_compose_path.is_file() and not docker_compose_path.is_dir():
                    found = False
                    break
            if found:
                return basedir
        return basedirs[-1]  # Fallback to last one

    @pytest.fixture(scope="session")
    def all_docker_projects(self):
        """
        Contains all docker projects for the current session, as a dict.
        """
        return {}

    @classmethod
    def get_docker_project(cls, request, all_docker_projects):
        """
        Get the Docker project, creating a new one if it doesn't exists.

        Returns the project instance, which can be used to start and stop the Docker containers.
        """
        testdir = request.fspath.dirname
        project_dir = request.config.getoption("project_dir")

        if project_dir:
            basedirs = [project_dir]
        else:
            basedirs = [d for d in (os.path.join(testdir, 'docker-compose'), testdir, '.') if Path(d).is_dir()]

        project = None
        for basedir in basedirs:
            try:
                project = all_docker_projects[basedir]
                break
            except KeyError:
                pass

        if not project:
            docker_compose_options = request.config.getoption("docker_compose")
            if docker_compose_options:
                files = []

                split_deprecated_displayed = False

                splitted_docker_compose_options = []
                for docker_compose_option in docker_compose_options:
                    if ',' in docker_compose_option:
                        if not split_deprecated_displayed:
                            warnings.warn(DeprecationWarning(
                                "Using ',' in --docker-compose option to specify multiple compose files is deprecated."
                                "You can now use --docker-compose many times in the same command."))
                            split_deprecated_displayed = True

                        for splitted_docker_compose_option in docker_compose_option.split(','):
                            splitted_docker_compose_options.append(splitted_docker_compose_option)
                    else:
                        splitted_docker_compose_options.append(docker_compose_option)

                if not project_dir:
                    project_dir = cls.projectdir_from_basedirs_and_docker_compose_options(
                        basedirs, docker_compose_options)

                for docker_compose_option in docker_compose_options:
                    docker_compose_path = Path(project_dir).joinpath(docker_compose_option)

                    if docker_compose_path.is_dir():
                        docker_compose_path = docker_compose_path.joinpath("docker-compose.yml")

                    if not docker_compose_path.is_file():
                        raise ValueError(
                            "Unable to find `{docker_compose}` "
                            "for integration tests in following directories: `{basedirs}`.".format(
                                docker_compose=docker_compose_option,
                                basedirs=basedirs
                            ),
                        )

                    files.append(str(docker_compose_path.relative_to(project_dir)))

                project_key = '|'.join(files)
                project = project_from_options(
                    project_dir=str(project_dir),
                    options={"--file": files},
                )
            else:
                project, project_key = cls.project_from_options_for_each_dir(basedirs, options={})

            all_docker_projects[project_key] = project

        if not request.config.getoption("--docker-compose-no-build"):
            project.build()

        if request.config.getoption("--use-running-containers"):
            if not request.config.getoption("--docker-compose-no-build"):
                warnings.warn(UserWarning(
                    "You used the '--use-running-containers' without the "
                    "'--docker-compose-no-build' flag, the newly build "
                    "containers won't be used if there are already "
                    "containers running!"))
            current_containers = project.containers()
            containers = project.up()
            if not set(current_containers) == set(containers):
                warnings.warn(UserWarning(
                    "You used the '--use-running-containers' but "
                    "pytest-docker-compose could not find all containers "
                    "running. The remaining containers have been started."))
        else:
            if any(project.containers()):
                raise ContainersAlreadyExist(
                    "There are already existing containers, please remove all "
                    "containers by running 'docker-compose down' before using "
                    "the pytest-docker-compose plugin. Alternatively, you "
                    "can use the '--use-running-containers' flag to indicate "
                    "you will use the currently running containers.")
        return project

    @classmethod
    def generate_scoped_containers_fixture(cls, scope: str):
        """
        Create scoped fixtures that retrieve or spin up all containers, and add
        network info objects to containers and then yield the containers for
        duration of test.

        After the tests wrap up the fixture prints the logs of each containers
        and tears them down unless '--use-running-containers' was supplied.
        """

        @pytest.fixture(scope=scope)  # type: ignore
        def scoped_containers_fixture(request, all_docker_projects):
            docker_project = cls.get_docker_project(request, all_docker_projects)

            now = datetime.utcnow()
            if request.config.getoption("--use-running-containers"):
                containers = docker_project.containers()  # type: List[Container]
            else:
                if any(docker_project.containers()):
                    raise ContainersAlreadyExist(
                        'pytest-docker-compose tried to start containers but there are'
                        ' already running containers: %s, you probably scoped your'
                        ' tests wrong' % docker_project.containers())
                containers = docker_project.up()
                if not containers:
                    raise ValueError("`docker-compose` didn't launch any containers!")

            container_getter = ContainerGetter(docker_project)
            yield container_getter

            if request.config.getoption("--verbose"):
                for container in sorted(containers, key=lambda c: c.name):
                    header = "Logs from {name}:".format(name=container.name)
                    print(header, '\n', "=" * len(header))
                    print(container.logs(since=now).decode("utf-8", errors="replace")
                          or "(no logs)", '\n')

            if not request.config.getoption("--use-running-containers"):
                docker_project.down(ImageType.none, request.config.getoption("--docker-compose-remove-volumes"))

        scoped_containers_fixture.__wrapped__.__doc__ = """
            Spins up the containers for the Docker project and returns an
            object that can retrieve the containers. The returned containers
            all have one additional attribute called network_info to simplify
            accessing the hostnames and exposed port numbers for each container.
            This set of containers is scoped to '%s'
            """ % scope
        return scoped_containers_fixture


plugin = DockerComposePlugin()


class ContainerGetter:
    """
    A class that retrieves containers from the docker project and adds a
    convenience wrapper for the available ports
    """

    def __init__(self, docker_project: Project) -> None:
        self.docker_project = docker_project

    def get(self, key: str, wait_running=True, timeout=15) -> Container:
        containers = self.docker_project.containers(service_names=[key], stopped=True)
        if not containers:
            raise ContainerDoesntExist("The service '%s' doesn't exists" % key)

        container = containers[0]
        if wait_running:
            start = time.time()
            while not container.is_running or container.is_restarting:
                if time.time() - start >= timeout:
                    raise ContainerNotRunning("The service '%s' is still %s after %s seconds" %
                                              (key, container.human_readable_state, timeout))
                time.sleep(0.1)
                containers = self.docker_project.containers(service_names=[key], stopped=True)
                if not containers:
                    raise ContainerDoesntExist("The service '%s' doesn't exists" % key)
                container = containers[0]

            # network_info is added only for wait running as it may lead to race condition issues.
            container.network_info = create_network_info_for_container(container)
        return container



# -*- coding: utf-8 -*-
import os
import re
from typing import Dict

from jinja2 import Template

from ddb.config import config
from ddb.feature.traefik.schema import ExtraServiceSchema
from ...action import Action, InitializableAction
from ...cache.removal import RemovalCacheSupport
from ...config.migrations import MigrationsDotty
from ...context import context
from ...event import events
from ...utils.file import write_if_different, copy_if_different, force_remove, FileUtils


def get_template(template: str) -> Template:
    """
    Retrieve the template object base on the input template string
    :param template: the template
    :return:
    """
    if re.compile('^(https?|file)://').match(template):
        return Template(FileUtils.get_file_content(template))
    return Template(template)


class TraefikInstalllCertsAction(Action):
    """
    Install generated certificates on traefik.
    """

    @property
    def event_bindings(self):
        return events.certs.available

    @property
    def name(self) -> str:
        return "traefik:install-certs"

    @staticmethod
    def execute(domain: str, private_key: str, certificate: str, wildcard: bool = False):
        """
        Execute action
        """
        certs_directory = config.data.get('traefik.certs_directory')
        mapped_certs_directory = config.data.get('traefik.mapped_certs_directory')
        config_directory = config.data.get('traefik.config_directory')

        private_key_filename = "%s.key" % (domain,)
        private_key_filename_target = os.path.join(certs_directory, private_key_filename)
        copy_if_different(private_key, private_key_filename_target, log=True)

        certificate_filename = "%s.crt" % (domain,)
        certificate_filename_target = os.path.join(certs_directory, certificate_filename)
        copy_if_different(certificate, certificate_filename_target, log=True)

        config_data = dict(config.data)
        config_data['_local'] = {
            'certFile': '/'.join([mapped_certs_directory, certificate_filename]),
            'keyFile': '/'.join([mapped_certs_directory, private_key_filename])
        }

        ssl_config = get_template(config.data.get('traefik.ssl_config_template')).render(config_data)

        config_target = os.path.join(config_directory, "%s.ssl.toml" % (domain,))
        if write_if_different(config_target, ssl_config, 'r', 'w'):
            context.log.success("SSL Configuration file written for domain %s" % (domain,))
        else:
            context.log.notice("SSL Configuration file exists for domain %s" % (domain,))


class TraefikUninstalllCertsAction(Action):
    """
    Uninstall removed certificates from traefik.
    """

    @property
    def event_bindings(self):
        return events.certs.removed

    @property
    def name(self) -> str:
        return "traefik:uninstall-certs"

    @staticmethod
    def execute(domain: str):
        """
        Execute action
        """
        config_directory = config.data.get('traefik.config_directory')
        config_target = os.path.join(config_directory, "%s.ssl.toml" % (domain,))

        if os.path.exists(config_target):
            force_remove(config_target)
            context.log.notice("SSL Configuration file removed for domain %s" % (domain,))

        certs_directory = config.data.get('traefik.certs_directory')
        private_key_filename = "%s.key" % (domain,)
        private_key_filename_target = os.path.join(certs_directory, private_key_filename)
        if os.path.exists(private_key_filename_target):
            force_remove(private_key_filename_target)
            context.log.notice("SSL private key file removed for domain %s" % (domain,))

        certificate_filename = "%s.crt" % (domain,)
        certificate_filename_target = os.path.join(certs_directory, certificate_filename)
        if os.path.exists(certificate_filename_target):
            force_remove(certificate_filename_target)
            context.log.notice("SSL certificate file removed for domain %s" % (domain,))


class TraefikExtraServicesAction(InitializableAction):
    """
    Generates extra services configuration to register any http/https URL inside the traefik reverse proxy.
    """

    def __init__(self):
        super().__init__()
        self.removal_cache_support = None  # type: RemovalCacheSupport

    @property
    def event_bindings(self):
        return events.phase.configure

    @property
    def name(self) -> str:
        return "traefik:extra-services"

    def initialize(self):
        self.removal_cache_support = RemovalCacheSupport("traefik.extra-services", {'generated-files', 'certs-domains'})

    def destroy(self):
        if self.removal_cache_support:
            self.removal_cache_support.close()
            self.removal_cache_support = None

    @staticmethod
    def get_extra_services():
        """
        Get extra services defined in configuration
        """
        extra_services = config.data.get('traefik.extra_services')  # type: Dict[str, ExtraServiceSchema]
        if extra_services:
            for id_, extra_service in extra_services.items():
                extra_service_data, config_data = \
                    TraefikExtraServicesAction._prepare_extra_service_data(extra_service, id_)
                yield id_, extra_service_data, config_data

    def execute(self):
        """
        Execute action
        """
        self.removal_cache_support.prepare()

        config_directory = config.data.get('traefik.config_directory')
        config_template = get_template(config.data.get('traefik.extra_services_config_template'))

        for id_, extra_service_data, config_data in TraefikExtraServicesAction.get_extra_services():
            rendered_config = config_template.render(config_data)

            if extra_service_data.get('domain'):
                prefix = extra_service_data.get('domain')
            else:
                prefix = '.'.join(
                    x for x in ['rule', config.data.get('core.domain.sub'), config.data.get('core.domain.ext')] if x
                )

            config_target = os.path.join(config_directory, "%s.extra-service.%s.toml" % (prefix, id_))
            if write_if_different(config_target, rendered_config, 'r', 'w'):
                context.log.success(
                    "Extra service configuration file written for domain %s" % (prefix,))
            else:
                context.log.notice(
                    "Extra service configuration file exists for domain %s" % (prefix,))

            self.removal_cache_support.set_current_value('generated-files', config_target)

            if extra_service_data.get('https') is not False:
                events.certs.generate(extra_service_data['domain'])
                self.removal_cache_support.set_current_value('certs-domains', extra_service_data['domain'])

        for (key, value) in self.removal_cache_support.get_removed():
            if key == 'certs-domains':
                events.certs.remove(value)
            elif key == 'generated-files':
                force_remove(value)

    @staticmethod
    def _prepare_extra_service_data(extra_service, id_):
        data = MigrationsDotty(dict(config.data))
        extra_service_data = dict(extra_service)
        extra_service_data['id'] = id_
        data['_local'] = extra_service_data
        if extra_service_data.get('domain'):
            extra_service_data['domain'] = Template(extra_service_data.get('domain')).render(data)
        if extra_service_data.get('rule'):
            extra_service_data['rule'] = Template(extra_service_data.get('rule')).render(data)
        if extra_service_data.get('url'):
            extra_service_data['url'] = Template(extra_service_data.get('url')).render(data)

        if extra_service['https'] is not None:
            extra_service_data['redirect_to_https'] = False
        if extra_service_data.get('redirect_to_https'):
            extra_service['https'] = True
        if extra_service_data.get('redirect_to_https') is None and extra_service['https'] is None:
            extra_service_data['redirect_to_https'] = config.data.get('jsonnet.docker.virtualhost.redirect_to_https')

        if not extra_service_data.get('path_prefix'):
            extra_service_data['redirect_to_path_prefix'] = False
        if extra_service['redirect_to_path_prefix'] is None:
            extra_service_data['redirect_to_path_prefix'] = \
                config.data.get('jsonnet.docker.virtualhost.redirect_to_path_prefix')

        return extra_service_data, data



import os

from ddb.__main__ import main
from tests.utilstest import expect_gitignore


class TestCopy:
    def test_copy(self, project_loader):
        project_loader("copy")

        main(["configure"])

        main(["activate"])

        assert os.path.exists("destination/test1")
        assert os.path.exists("destination/test2")

        assert expect_gitignore('.gitignore', '/destination/test1', '/destination/test2')

        os.remove("destination/test1")

        main(["configure"])

        assert os.path.exists("destination/test1")
        assert os.path.exists("destination/test2")
        assert expect_gitignore('.gitignore', '/destination/test1', '/destination/test2')



from typing import Iterable, Union, TYPE_CHECKING

from dotty_dict import Dotty
from marshmallow import Schema
from marshmallow.fields import Nested, Dict, List

if TYPE_CHECKING:
    from ddb.feature import Feature


def _get_stop_fields_from_schema(schema: Schema, stack, ret):
    for field_name, field in schema.fields.items():
        stack.append(field_name)
        if isinstance(field, Dict):
            ret.append(tuple(stack))
        if isinstance(field, List):
            ret.append(tuple(stack))
        if isinstance(field, Nested):
            _get_stop_fields_from_schema(field.schema, stack, ret)
        stack.pop()


def _get_stop_fields(features: Iterable['Feature']):
    ret = []
    stack = []
    for feature in features:
        stack.append(feature.name)
        _get_stop_fields_from_schema(feature.schema(), stack, ret)
        stack.pop()

    return ret


def to_environ(data: Union[Dotty, dict], env_prefix) -> dict:
    """
    Export configuration to environment dict.
    """
    return _flatten(env_prefix, "_", "_%s_",
                    lambda x: str.upper(x).replace('-', '_'),
                    str,
                    data=dict(data))


def flatten(data: Union[Dotty, dict], prefix=None, sep=".", array_index_format="[%s]",
            key_transformer=None, value_transformer=None, keep_primitive_list=False,
            stop_for_features=None) -> dict:
    """
    Export configuration to a flat dict.
    """
    stop_for = tuple(map(sep.join, _get_stop_fields(stop_for_features))) if stop_for_features is not None else ()

    return _flatten(prefix, sep, array_index_format,
                    key_transformer, value_transformer,
                    keep_primitive_list, stop_for, data=dict(data))


def _flatten(prefix=None, sep=".", array_index_format="[%s]",
             key_transformer=None, value_transformer=None, keep_primitive_list=False,
             stop_for=(), data=None, output=None) -> dict:
    if output is None:
        output = {}

    if prefix is None:
        prefix = ""
    if key_transformer is None:
        key_transformer = lambda x: x
    if value_transformer is None:
        value_transformer = lambda x: x

    stop_recursion = False
    if prefix in stop_for:
        stop_recursion = True

    if not stop_recursion and isinstance(data, dict):
        for (name, value) in data.items():
            key_prefix = (prefix + sep if prefix else "") + key_transformer(name)
            key_prefix = key_transformer(key_prefix)

            _flatten(key_prefix, sep, array_index_format,
                     key_transformer, value_transformer, keep_primitive_list,
                     stop_for, value, output)

    elif not stop_recursion and isinstance(data, list):
        if keep_primitive_list and (not data or
                                    not set(filter(lambda x: x not in (int, float, bool, str), set(map(type, data))))):
            output[prefix] = value_transformer(data)
        else:
            i = 0
            for value in data:
                replace_prefix = (prefix if prefix else "") + (array_index_format % str(i))
                replace_prefix = key_transformer(replace_prefix)

                _flatten(replace_prefix, sep, array_index_format,
                         key_transformer, value_transformer, keep_primitive_list,
                         stop_for, value, output)

                i += 1
    else:
        output[prefix] = value_transformer(data)

    return output



from abc import ABC, abstractmethod
from typing import Optional, Set

from dotty_dict import Dotty

from ddb.config.merger import config_merger
from ddb.context import context

silent = False
_warns = set()


class AbstractDeprecation(ABC):
    """
    Abstract deprecation.
    """

    def warn(self, source=None):
        """
        Display deprecated warning.
        """


class AbstractMigration(AbstractDeprecation):
    """
    Abstract migration.
    """

    @abstractmethod
    def verify(self, config: Dotty) -> bool:
        """
        Verify the given configuration with this migration.
        :param config: loaded configuration
        :return: True if migration should be applied, else otherwise
        """

    @abstractmethod
    def migrate(self, config: Dotty):
        """
        Apply changes on configuration to make it work with latest changes.
        :param config: loaded configuration
        """

    @abstractmethod
    def get_new_value(self, config: Dotty):
        """
        Get property new value, regardless it's defined the new way or the old way.
        :param config: loaded configuration
        :raise KeyError: If no value is found in config.
        """

    @abstractmethod
    def get_old_value(self, config: Dotty):
        """
        Get property old value, regardless it's defined the new way or the old way.
        :param config: loaded configuration
        :raise KeyError: If no value is found in config.
        """


class AbstractPropertyMigration(AbstractMigration):
    """
    Abstract migration based on single move/rename.
    """

    @property
    @abstractmethod
    def old_config_key(self) -> str:
        """
        Old property configuration key
        """

    @property
    @abstractmethod
    def new_config_key(self) -> str:
        """
        New property configuration key
        """

    @property
    @abstractmethod
    def requires_value_migration(self) -> bool:
        """
        Does this migration change the value of configuration when migrating to new config key.
        """


class PropertyMigration(AbstractPropertyMigration):
    """
    Migration based on single move/rename. Value can be changed using an optional transformer function.
    """

    def __init__(self,
                 old_config_key,
                 new_config_key,
                 transformer=None,
                 rollback_transformer=None,
                 since=None,
                 removed_in="a future major release"):
        self._default_transformer = lambda value, config: value
        self._old_config_key = old_config_key
        self._new_config_key = new_config_key
        self.transformer = transformer if transformer else self._default_transformer
        self.rollback_transformer = rollback_transformer if rollback_transformer else self._default_transformer
        self.since = since
        self.removed_in = removed_in

    @property
    def old_config_key(self):
        return self._old_config_key

    @property
    def new_config_key(self):
        return self._new_config_key

    @property
    def requires_value_migration(self):
        return self.transformer != self._default_transformer

    def verify(self, config: Dotty):
        if self.old_config_key in config:
            self.warn()
            return False
        return True

    def warn(self, source=None):
        if not silent:
            since_message = f" since {self.since}" if self.since else ""
            source_message = f" (in \"{source}\")." if source else "."

            message = "[deprecated] " + \
                      f"\"{self.old_config_key}\" configuration property is deprecated{since_message} " \
                      f"and will be removed in {self.removed_in}. " + \
                      f"You should use \"{self.new_config_key}\" instead{source_message}"

            if message not in _warns:
                _warns.add(message)
                context.log.warning(message)

    def migrate(self, config: Dotty):
        old_value = config.pop(self.old_config_key)
        if old_value is not None:
            transformed_value = self.transformer(old_value, config)
            config[self.new_config_key] = transformed_value

    def get_new_value(self, config: Dotty):
        if self.new_config_key in config:
            return config.get(self.new_config_key)
        if self.old_config_key in config:
            return self.transformer(config.get(self.old_config_key), config)
        raise KeyError

    def get_old_value(self, config: Dotty):
        if self.new_config_key in config:
            return self.rollback_transformer(config.get(self.new_config_key), config)
        if self.old_config_key in config:
            return config.get(self.old_config_key)
        raise KeyError


class MigrationsDotty(Dotty):
    """
    Dotty that supports migrations when requesting deprecated keys and displays deprecation warnings.
    """

    def __init__(self, dictionary=None, namespace=""):
        if dictionary is None:
            dictionary = {}
        self.namespace = namespace
        super().__init__(dictionary)

    def _build_deprecation_dict(self, item):
        migrations = get_migrations_from_old_config_key_startswith(item + ".")
        deprecation_dict = Dotty({})

        for migration in migrations:
            try:
                deprecation_dict[migration.old_config_key[len(item + "."):]] = migration.get_new_value(self)
            except KeyError:
                pass

        return dict(deprecation_dict)

    def raw(self):
        """
        Build a new default Dotty instance using same data.
        :return:
        """
        return Dotty(self._data)

    def __getitem__(self, item):
        if isinstance(item, str):
            property_migration = get_migration_from_old_config_key(item)
            if property_migration:
                if not silent:
                    property_migration.warn()
                try:
                    return property_migration.get_old_value(Dotty(self._data))
                except KeyError:
                    pass
        try:
            value = super().__getitem__(item)
        except KeyError as key_error:
            deprecation_dict = self._build_deprecation_dict(item)
            if deprecation_dict:
                return deprecation_dict
            raise key_error

        if isinstance(value, dict):
            deprecation_dict = self._build_deprecation_dict(item)
            if deprecation_dict:
                value = config_merger.merge(deprecation_dict, value)

        return value


_default_history = (
    PropertyMigration("docker.build_image_tag_from_version",
                      "jsonnet.docker.build.image_tag_from", since="v1.6.0"),
    PropertyMigration("docker.build_image_tag_from",
                      "jsonnet.docker.build.image_tag_from", since="v1.6.0"),
    PropertyMigration("docker.build_image_tag",
                      "jsonnet.docker.build.image_tag", since="v1.6.0"),

    PropertyMigration("docker.cache_from_image",
                      "jsonnet.docker.build.cache_from_image", since="v1.6.0"),
    PropertyMigration("docker.directory",
                      "jsonnet.docker.build.context.base_directory", since="v1.6.0"),

    PropertyMigration("docker.jsonnet.binary_disabled",
                      "jsonnet.docker.binary.disabled", since="v1.6.0"),
    PropertyMigration("docker.jsonnet.virtualhost_disabled",
                      "jsonnet.docker.virtualhost.disabled", since="v1.6.0"),

    PropertyMigration("docker.reverse_proxy.network_id",
                      "jsonnet.docker.virtualhost.network_id", since="v1.6.0"),
    PropertyMigration("docker.reverse_proxy.network_names",
                      "jsonnet.docker.networks.names", since="v1.6.0"),
    PropertyMigration("docker.reverse_proxy.certresolver",
                      "jsonnet.docker.virtualhost.certresolver", since="v1.6.0"),
    PropertyMigration("docker.reverse_proxy.https",
                      "jsonnet.docker.virtualhost.https", since="v1.6.0"),
    PropertyMigration("docker.reverse_proxy.redirect_to_https",
                      "jsonnet.docker.virtualhost.redirect_to_https", since="v1.6.0"),
    PropertyMigration("docker.reverse_proxy.redirect_to_path_prefix",
                      "jsonnet.docker.virtualhost.redirect_to_path_prefix", since="v1.6.0"),

    PropertyMigration("docker.debug.disabled",
                      "jsonnet.docker.xdebug.disabled", since="v1.6.0"),
    PropertyMigration("docker.debug.host",
                      "jsonnet.docker.xdebug.host", since="v1.6.0"),

    PropertyMigration("docker.compose.service_context_root",
                      "jsonnet.docker.build.context.use_project_home", since="v1.6.0"),

    PropertyMigration("docker.compose.service_init",
                      "jsonnet.docker.service.init", since="v1.6.0"),

    PropertyMigration("docker.user.name",
                      "jsonnet.docker.user.name", since="v1.6.0"),
    PropertyMigration("docker.user.group",
                      "jsonnet.docker.user.group", since="v1.6.0"),
    PropertyMigration("docker.user.name_to_uid",
                      "jsonnet.docker.user.name_to_uid", since="v1.6.0"),
    PropertyMigration("docker.user.group_to_gid",
                      "jsonnet.docker.user.group_to_gid", since="v1.6.0"),

    PropertyMigration("docker.registry.name",
                      "jsonnet.docker.registry.name", since="v1.6.0"),
    PropertyMigration("docker.registry.repository",
                      "jsonnet.docker.registry.repository", since="v1.6.0"),

    PropertyMigration("docker.compose.project_name",
                      "jsonnet.docker.compose.project_name", since="v1.6.0"),
    PropertyMigration("docker.compose.network_name",
                      "jsonnet.docker.compose.network_name", since="v1.6.0"),
    PropertyMigration("docker.compose.file_version",
                      "jsonnet.docker.compose.version", since="v1.6.0"),

    PropertyMigration("docker.disabled_services",
                      "jsonnet.docker.compose.excluded_services", since="v1.6.0"),

    PropertyMigration("docker.build_image_tag",
                      "jsonnet.docker.build.image_tag", since="v1.6.0"),
    PropertyMigration("docker.build_image_tag",
                      "jsonnet.docker.build.image_tag_from", since="v1.6.0"),

    PropertyMigration("docker.port_prefix",
                      "jsonnet.docker.expose.port_prefix", since="v1.6.0"),
    PropertyMigration("docker.restart_policy",
                      "jsonnet.docker.service.restart", since="v1.6.0"),

    PropertyMigration("docker.reverse_proxy.type",
                      "jsonnet.docker.virtualhost.disabled", since="v1.6.0",
                      transformer=lambda type, config:
                      type == 'none' or (type is not None and not type),
                      rollback_transformer=lambda disabled, config:
                      config.get('jsonnet.docker.virtualhost.type') if not disabled else 'none'),
)

_history = None
_history_from_old_config_key_dict = None
_history_from_new_config_key_dict = None


def get_history():
    """
    Get migrations history.
    :return:
    """
    global _history  # pylint:disable=global-statement,global-variable-not-assigned
    return _history


def set_history(history=_default_history):
    """
    Set migrations history.
    """
    global _history, _history_from_old_config_key_dict, _history_from_new_config_key_dict  # pylint:disable=global-statement
    _history = history

    _history_from_old_config_key_dict = {
        migration.old_config_key: migration for migration in history
        if isinstance(migration, AbstractPropertyMigration) and migration.old_config_key
    }

    _history_from_new_config_key_dict = {
        migration.new_config_key: migration for migration in history
        if isinstance(migration, AbstractPropertyMigration) and migration.new_config_key
    }
    _warns.clear()


set_history()


def get_migration_from_old_config_key(old_config_key: str) -> Optional[AbstractPropertyMigration]:
    """
    Get a migration from old config key.
    """
    return _history_from_old_config_key_dict.get(old_config_key)


def get_migration_from_new_config_key(new_config_key: str) -> Optional[AbstractPropertyMigration]:
    """
    Get a migration from new config key.
    """
    return _history_from_new_config_key_dict.get(new_config_key)


def get_migrations_from_old_config_key_startswith(old_config_key_start: str) -> Set[AbstractPropertyMigration]:
    """
    Get all migrations where old_config_key starts with given value
    """
    ret = set()
    for migration in get_history():
        if isinstance(migration, AbstractPropertyMigration) and \
                migration.old_config_key and \
                migration.old_config_key.startswith(old_config_key_start):
            ret.add(migration)
    return ret


def migrate(config: Dotty):
    """
    Migrate all defined migrations in given config data.
    """
    global silent  # pylint:disable=global-statement
    old_silent = silent
    try:
        silent = True
        for migration in get_history():
            if not migration.verify(config):
                migration.migrate(config)
    finally:
        silent = old_silent



# -*- coding: utf-8 -*-
"""
Crypto module
"""

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization


class CryptoException(Exception):
    """
    Exception raised when something wrong occurs in cryto module.
    """


def convert_pem_to_der(part, content):
    """
    Converts a PEM certificate, csr or crl to DER
    :param part: type of object to convert
    :type part: byte[]
    :param content: PEM data as a ASCII string
    :type content: str
    :return: DER Encoded object
    :rtype: byte[]
    """
    if part == 'certificate':
        cert = x509.load_pem_x509_certificate(content, default_backend())
        return cert.public_bytes(serialization.Encoding.DER)
    if part == 'certificate_request':
        csr = x509.load_pem_x509_csr(content, default_backend())
        return csr.public_bytes(serialization.Encoding.DER)
    raise CryptoException('Unknown object type: %s' % part)



# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from argparse import ArgumentParser
from typing import Iterable, Union, Optional, Callable, Any

from ..context import context
from ..phase import phases, Phase
from ..phase.phase import execute_phase
from ..registry import RegistryObject, DefaultRegistryObject


class Command(RegistryObject, ABC):
    """
    A command is available in the program usage and can perform some action on the system.
    """

    @property
    @abstractmethod
    def allow_unknown_args(self):
        """
        Check if this command allow unknown args.
        """

    @property
    @abstractmethod
    def avoid_stdout(self):
        """
        Should this command avoid stdout spoiling
        """

    @abstractmethod
    def execute(self):
        """
        Execute the command.
        """

    def configure_parser(self, parser: ArgumentParser):
        """
        Configure the argument parser.
        """

    def add_parser(self, subparsers):
        """
        Add command parser into given subparsers.
        """
        return subparsers.add_parser(self.name, help=self.description)


class DefaultCommand(DefaultRegistryObject, Command):
    """
    A command is available in the program usage and can perform some action on the system.
    """

    def __init__(self, name: str, description: str = None, parent: Optional[Union[Command, str]] = None,
                 allow_unknown_args=False, before_execute: Optional[Callable[[], Any]] = None, avoid_stdout=False):
        super().__init__(name, description)
        self.before_execute = before_execute
        self._parent = parent
        self._allow_unknown_args = allow_unknown_args
        self._avoid_stdout = avoid_stdout

    @property
    def allow_unknown_args(self):
        return self._allow_unknown_args

    @property
    def avoid_stdout(self):
        return self._avoid_stdout

    @property
    def parent(self) -> Optional[Command]:
        """
        Parent command
        """
        if not self._parent:
            return None

        if isinstance(self._parent, Command):
            return self._parent

        # Inline import because of recursive imports if declared in module
        from ..command import commands  # pylint:disable=import-outside-toplevel,cyclic-import
        command = commands.get(self._parent)
        if not command:
            raise ValueError
        return command

    def execute(self):
        """
        Execute the command.
        """
        if self.before_execute:
            self.before_execute()

        if self.parent:
            self.parent.execute()


class LifecycleCommand(DefaultCommand):
    """
    A command that will emit events based on lifecycle phases.
    Triggered events are named "phase:<phase.name>"
    """

    def __init__(self, name: str, description: Union[str, None], *lifecycle: Union[str, Phase],
                 parent: Optional[Union[Command, str]] = None, before_execute: Optional[Callable[[], Any]] = None,
                 avoid_stdout=False):
        super().__init__(name, description, parent, before_execute=before_execute, avoid_stdout=avoid_stdout)
        self._lifecycle = list(map(lambda phase: phases.get(phase) if not isinstance(phase, Phase) else phase,
                                   lifecycle))  # type: Iterable[Phase]
        for phase in self._lifecycle:
            if phase.allow_unknown_args:
                self._allow_unknown_args = True

    def configure_parser(self, parser: ArgumentParser):
        super().configure_parser(parser)
        for phase in self._lifecycle:
            phase.configure_parser(parser)

    def execute(self):
        super().execute()
        for phase in self._lifecycle:
            execute_phase(phase)


class LifecycleWithoutHelpCommand(LifecycleCommand):
    """
    A LifecycleCommand that doesn't add --help option to it's parser.
    """

    def add_parser(self, subparsers):
        """
        Add command parser into given subparsers.
        """
        return subparsers.add_parser(self.name, help=self.description, add_help=False)


def execute_command(command: Command):
    """
    Execute a command with context update.
    """

    context.command = command
    try:
        command.execute()
    finally:
        context.command = None



# -*- coding: utf-8 -*-
import os
import platform
from argparse import ArgumentParser
from typing import Iterable, ClassVar

import distro
from dotty_dict import Dotty

from .actions import FeaturesAction, ConfigAction, ReloadConfigAction, EjectAction, SelfUpdateAction, \
    CheckForUpdateAction, VersionAction, CheckRequiredVersion
from .schema import CoreFeatureSchema
from ..feature import Feature, FeatureConfigurationAutoConfigureError, FeatureConfigurationReadOnlyError
from ..schema import FeatureSchema
from ...action import Action
from ...action.runner import ExpectedError, FailFastError
from ...command import LifecycleCommand, Command
from ...config import config
from ...config.config import ConfigPaths
from ...context import context
from ...phase import Phase, DefaultPhase


class ConfigureSecondPassException(Exception):
    """
    Exception that should be raised when an additional configuration file is to be loaded.
    """


class NoProjectConfigurationError(FailFastError, ExpectedError):
    """
    Error that should be raised when a project configuration file is required for the command.
    """

    def __init__(self):
        super().__init__("No project configuration file found. "
                         "Please create a ddb.yml file in your project directory. "
                         "It can be empty.")

    def log_error(self):
        context.log.error(str(self))


class CoreFeature(Feature):
    """
    Default commands and configuration support.
    """

    @property
    def name(self) -> str:
        return "core"

    @property
    def schema(self) -> ClassVar[FeatureSchema]:
        return CoreFeatureSchema

    @property
    def actions(self) -> Iterable[Action]:
        return (
            FeaturesAction(),
            ConfigAction(),
            ReloadConfigAction(),
            EjectAction(),
            SelfUpdateAction(),
            CheckForUpdateAction(),
            VersionAction(),
            CheckRequiredVersion()
        )

    @property
    def phases(self) -> Iterable[Phase]:
        def configure_parser(parser: ArgumentParser):
            parser.add_argument("--eject", action="store_true",
                                help="Eject the project using the current configuration")
            parser.add_argument("--autofix", action="store_true",
                                help="Autofix supported deprecated warnings by modifying template sources.")

        def config_parser(parser: ArgumentParser):
            parser.add_argument("property", nargs='?',
                                help="Property to read")
            parser.add_argument("--variables", action="store_true",
                                help="Output as a flat list of variables available in template engines")
            parser.add_argument("--value", action="store_true",
                                help="Output value of given property")
            parser.add_argument("--full", action="store_true",
                                help="Output full configuration")
            parser.add_argument("--files", action="store_true",
                                help="Group by loaded configuration file")

        def info_parser(parser: ArgumentParser):
            parser.add_argument("--type", action="append",
                                help="Filter for a type of information between: bin, env, port and vhost")

        def selfupdate_parser(parser: ArgumentParser):
            parser.add_argument("--force", action="store_true", help="Force update")

        return (
            DefaultPhase("init", "Initialize project", run_once=True),
            DefaultPhase("configure", "Configure the environment", configure_parser),
            DefaultPhase("download", "Download files from remote sources"),
            DefaultPhase("features", "Display enabled features"),
            DefaultPhase("config", "Display effective configuration", config_parser),
            DefaultPhase("info", "Display useful information", info_parser),
            DefaultPhase("selfupdate", "Update ddb binary with latest version", parser=selfupdate_parser)
        )

    @property
    def commands(self) -> Iterable[Command]:
        def requires_project_config():
            if not config.project_configuration_file:
                error = NoProjectConfigurationError()
                error.log_error()
                raise error

        return (
            LifecycleCommand("init", "Initialize the environment",
                             "init", avoid_stdout=True),

            LifecycleCommand("configure", "Configure the environment",
                             "configure",
                             parent="init",
                             before_execute=requires_project_config),

            LifecycleCommand("download", "Download files from remote sources",
                             "download"),

            LifecycleCommand("features", "List enabled features",
                             "features"),

            LifecycleCommand("config", "Display effective configuration",
                             "config"),

            LifecycleCommand("info", "Display useful information",
                             "info"),

            LifecycleCommand("self-update", "Update ddb to latest version",
                             "selfupdate"),
        )

    def configure(self, bootstrap=False):
        super().configure(bootstrap)
        if bootstrap:
            return
        self._load_environment_configuration()
        self._apply_eject_configuration()

    @staticmethod
    def _load_environment_configuration():
        """
        Loading enviromnent configuration file, if exists.
        """
        current = config.data.get('core.env.current')

        if config.filenames:
            filenames = list(config.filenames)
            original_filenames = list(config.filenames)

            current_env_filename = filenames[0] + '.' + current
            if current_env_filename not in filenames:
                filenames.insert(len(filenames) - 1, current_env_filename)

            extra = config.data.get('core.configuration.extra')
            if extra:
                for extra_item in extra:
                    if extra_item not in filenames:
                        filenames.insert(len(filenames) - 1, extra_item)

            if filenames != original_filenames:
                config.filenames = tuple(filenames)

                raise ConfigureSecondPassException()

    @staticmethod
    def _apply_eject_configuration():
        """
        Override some configuration that doesn't make sense without ddb
        :return:
        """
        if config.eject:
            config.data['core.path.ddb_home'] = '.ddb-home'
            config.data['core.path.home'] = '.docker-devbox-home'
            config.data['core.path.project_home'] = '.'

    def _configure_defaults(self, feature_config: Dotty):
        if not feature_config.get('project.name'):
            project_name = os.path.basename(config.paths.project_home)
            feature_config['project.name'] = project_name

        if not feature_config.get('domain.sub'):
            feature_config['domain.sub'] = feature_config['project.name'].replace("_", "-").replace(" ", "-")

        if feature_config.get('domain.value'):
            raise FeatureConfigurationReadOnlyError(self, 'domain.value')
        feature_config['domain.value'] = '.'.join((feature_config['domain.sub'], feature_config['domain.ext']))

        if not feature_config.get('env.current') and feature_config.get('env.available'):
            feature_config['env.current'] = feature_config['env.available'][-1]

        if not feature_config.get('env.current') or \
                feature_config.get('env.current') not in feature_config['env.available']:
            raise FeatureConfigurationAutoConfigureError(self, 'env.current')

        if not feature_config.get('path.project_home') and config.paths.project_home:
            feature_config['path.project_home'] = config.paths.project_home

        if not feature_config.get('path.home') and config.paths.home:
            feature_config['path.home'] = config.paths.home

        if not feature_config.get('path.ddb_home') and config.paths.ddb_home:
            feature_config['path.ddb_home'] = config.paths.ddb_home

        self._configure_release_asset_name_defaults(feature_config)

        config.path = ConfigPaths(ddb_home=feature_config.get('path.ddb_home'),
                                  home=feature_config.get('path.home'),
                                  project_home=feature_config.get('path.project_home'))

    def _configure_release_asset_name_defaults(self, feature_config: Dotty):
        if not feature_config.get('release_asset_name'):
            feature_config['release_asset_name'] = self._get_default_binary_remote_name()

    @staticmethod
    def _get_default_binary_remote_name():
        """
        Get default binary remote name, based on current platform.
        """
        if platform.system() == 'Windows':
            return 'ddb-windows.exe'
        if platform.system() == 'Darwin':
            return 'ddb-macos'
        if platform.system() == 'Linux':
            if distro.id() == 'alpine':
                return 'ddb-alpine'
            return 'ddb-linux'
        return None
